###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

from mod_python import apache
from mod_python import util
from pyano import *


class NewsInterface(Interface):

    title = 'Pyano Anonymous Usenet Interface'
    help = 'To use this, enter the newsgroups you wish to post to (comma separated), the from line you wish to have on the message (it is important that you read the part about from lines below), enter the subject of the message, enter the references line if replying to a post and you want it to thread (input the message ID of the post you are responding to, including the brackets &lt;&gt;, eg. &lt;asasfasf22228888765444@somewhere.com&gt;), and select if you want google to archive it or not.'

    def validate(self):
        groups = str(self.fs['newsgroups']).replace(' ','') # filter out whitespaces
        val_newsgroups(groups)
        orig = str(self.fs['from'])
        if orig:
            val_email(orig)
        refs = str(self.fs['references'])
        if refs:
            val_references(refs)
        mail2news = str(self.fs['mail2news']).replace(' ','')
        if mail2news:
            val_mail2news(mail2news)
        hashcash = str(self.fs['hashcash']).strip()
        if hashcash:
            val_hashcash(hashcash)
        try:
            checked = str(self.fs['archive'])
            no_archive = True
        except KeyError:
            no_archive = False
        subj = str(self.fs['subject'])
        if not subj:
            raise InputError('Subject required.')
        chain = parse_chain(self.fs,self.stats)
        n_copies = int(self.fs['copies'])
        val_n_copies(n_copies)
        msg = str(self.fs['message'])
        if not msg:
            raise InputError('Refusing to send empty message.')
        return groups, orig, subj, refs, mail2news, hashcash, no_archive, chain, n_copies, msg


    def header(self):
        return html_from_file(conf.news_header)


    def form(self,req):
        req.write('''
<h2>Web Interface</h2>

<form action="'''+str(req.uri)+'''" method="post" >
<table id="mixtable">
  <tr>
    <td><strong>*Newsgroup(s):</strong></td>
    <td><input class="line" name="newsgroups" value="" /></td>
  </tr>
  <tr>
    <td><strong>From:</strong></td>
    <td><input class="line" name="from" value="" /></td>
  </tr>
  <tr>
    <td><strong>*Subject:</strong></td>
    <td><input class="line" name="subject" value="" /></td>
  </tr>
  <tr>
    <td><strong>References:</strong></td>
    <td><input class="line" name="references" value="" /></td>
  </tr>
  <tr>
    <td><strong>Mail2News:</strong></td>
    <td><input class="line" name="mail2news" value="'''+conf.mail2news+'''" /></td>
  </tr>
  <tr>
    <td><strong>Hashcash:</strong></td>
    <td><input class="line" name="hashcash" value="" /></td>
  </tr>
  <tr>
    <td><strong>X-No-Archive:</strong></td>
    <td><input type="checkbox" name="archive" checked="checked" />
        (checked means Google will not archive this post)
    </td>
  </tr>
  <tr>''')
        if self.stats:
            req.write('''   <td><strong>Chain:</strong></td>
    <td>
'''+html_chain(self.stats)+'''
    </td>
  </tr>''')
        req.write('''  <tr>
   <td><strong>Copies:</strong></td>
    <td>
'''+html_copies()+'''
    </td>
  </tr>

  <tr>
    <td><strong>*Message:</strong></td>
    <td><textarea name="message" rows="30" cols="70" ></textarea></td>
  </tr>
  <tr>
    <td></td>
    <td><input type="submit" value="Send" /><input type="reset" value="Reset" /></td>
  </tr>
</table>
</form>

''')
    
    def process(self):
        newsgroups, orig, subj, refs, mail2news, hashcash, no_archive, chain, n_copies, msg = self.validate()
        send_news(newsgroups, orig, subj, refs, mail2news, hashcash, no_archive, chain, n_copies, msg)
        msg = 'Successfully sent message to '+newsgroups+' using '
        if chain:
            msg += 'remailer chain '+','.join(chain)+'.'
        else:
            msg += 'a random remailer chain.'
        return msg

    

def handler(req):
    return NewsInterface()(req)



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

import re
from config import conf

LATENT = 0
UPTIME = 1
MIDDLE = 2
FROM = 3


def get_stats():
    stats = {}
    try:
        _read_mlist(stats)
    except IOError: # could not read mlist2 file
        return None
    try:    
        _read_allow_from(stats)
    except IOError: # could not read from.html file
        pass
    return stats


def _read_mlist(stats):
    with open(conf.mlist2,'r') as f:
        n = 0
        in_stats_block = False
        in_remailer_block = False
        for line in f:
            n += 1
            if n == 5:
                in_stats_block = True
            if n >= 5 and in_stats_block:
                elems = line.split()
                if len(elems) == 0: # blank line, we are out of the stats block
                    in_stats_block = False
                else:
                    name = elems[0]
                    latent = elems[2]
                    uptime = float(elems[4].strip('%'))
                    stats[name] = [latent, uptime]

            else: # we are now looking for the remailer capabilities block 
                if '=' in line:
                    in_remailer_block = True
                if in_remailer_block:
                    name = line[11:].split('"',1)[0]
                    if "middle" in line:
                        stats[name].append(True)
                    else:
                        stats[name].append(False)


def _read_allow_from(stats):
    # by default no remailers accept from headers
    for name, s in stats.iteritems():
        stats[name].append(False)

    with open(conf.allow_from,'r') as f: # find out those that do accept from headers
        in_from_block = False
        m = re.compile("<td>(\w+)</td>")
        for line in f:
            if line.find("User Supplied From") >= 0:
                in_from_block = True
            if in_from_block:
                ok = m.search(line)
                if ok:
                    name = ok.group(1)
                    if name in stats:
                        stats[name][FROM] = True


def format_stats(name, stats):
    latent, uptime, middle, orig = stats
    s = name.ljust(12)+latent.rjust(5)+(str(uptime)+"%").rjust(7)
    if middle:
        s += " middle"
    if orig:
        s += " from"
    return s


def uptime_sort(stats):
    remailers = stats.keys()
    remailers.sort(cmp=lambda x,y: cmp(stats[y][UPTIME],stats[x][UPTIME]))
    return remailers



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

from pyano import *

class MailInterface(Interface):
    
    title = 'Pyano Anonymous Email Web Interface'
    help = 'To use this, enter the address you wish to send e-mail, the from line you wish to have on the message (it is important that you read the part about from lines below), and the subject.'

    def validate(self):
        to = str(self.fs['to'])
        val_email(to)
        orig = str(self.fs['from'])
        if orig:
            val_email(orig)
        subj = str(self.fs['subject'])
        chain = parse_chain(self.fs,self.stats)
        n_copies = int(self.fs['copies'])
        val_n_copies(n_copies)
        msg = str(self.fs['message'])
        if not msg:
            raise InputError('Refusing to send empty message.')
        return to, orig, subj, chain, n_copies, msg


    def header(self):
        return html_from_file(conf.mail_header)
    

    def form(self,req):
        req.write('''
<h2>Web Interface</h2>

<form action="'''+str(req.uri)+'''" method="post" >
<table id="mixtable">
  <tr>
    <td><strong>*To:</strong></td>
    <td><input class="line" name="to" value="" /></td>
  </tr>
  <tr>
    <td><strong>From:</strong></td>
    <td><input class="line" name="from" value="" /></td>
  </tr>
  <tr>
    <td><strong>Subject:</strong></td>
    <td><input class="line" name="subject" value="" /></td>
  </tr>
  <tr>''')
        if self.stats:
            req.write('''   <td><strong>Chain:</strong></td>
    <td>
'''+html_chain(self.stats)+'''
    </td>
  </tr>''')
        req.write('''  <tr>
   <td><strong>Copies:</strong></td>
    <td>
'''+html_copies()+'''
    </td>
  </tr>

  <tr>
    <td><strong>*Message:</strong></td>
    <td><textarea name="message" rows="30" cols="70" ></textarea></td>
  </tr>
  <tr>
    <td></td>
    <td><input type="submit" value="Send" /><input type="reset" value="Reset" /></td>
  </tr>
</table>
</form>

''')
        
    def process(self):
        to, orig, subj, chain, n_copies, msg = self.validate() # check user input
        send_mail(to,orig,subj,chain,n_copies,msg) # try sending to mixmaster
        msg = 'Successfully sent message to '+to+' using '
        if chain:
            msg += 'remailer chain '+','.join(chain)+'.'
        else:
            msg += 'a random remailer chain.'
        return msg

    

def handler(req):
    return MailInterface()(req)




###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################


import subprocess
from config import conf

class MixError(Exception):
    pass

def send_mail(to,orig,subj,chain,n_copies,msg):
    args = [conf.mixmaster,'--mail','--to='+to,'--subject='+subj,'--copies='+str(n_copies)]
    
    if orig:
        args.append('--header=From: '+orig)
    if chain:
        args.append('--chain='+','.join(chain))
    send_mix(args,msg)


def send_news(newsgroups, orig, subj, refs, mail2news, hashcash, no_archive, chain, n_copies, msg):
    args = [conf.mixmaster,'--subject='+subj,'--copies='+str(n_copies)]
    if orig:
        args.append('--header=From: '+orig)
    if no_archive:
        args.append('--header=X-No-Archive: yes')
    if mail2news: # user supplied mail2news gateways
        args.extend(['--mail','--to='+mail2news])
        args.append('--header=Newsgroups: '+newsgroups)
    else: # no supplied mail2news gateways, use mixmaster's internal mail2news parameters
        args.extend(['--post','--post-to='+newsgroups])
    if refs:
        args.append('--header=References: '+refs)
    if hashcash:
        args.append('--header=X-HashCash: '+hashcash)
    if chain:
        args.append('--chain='+','.join(chain))
    send_mix(args,msg)
    

def send_mix(args,msg):
    try:
        mix = subprocess.Popen(args,stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        out,err = mix.communicate(msg)
        if err.find('Error') >= 0:
            raise MixError('Mixmaster process returned the following error: '+str(err)+'. Sending failed.')
    except OSError: # usally means that mixmaster could not be executed
        raise MixError('Could not find mixmaster binary.')



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

from mod_python import util
from mod_python import apache
from pyano import *

class Interface:
    
    def __call__(self,req):
        parse_config( req.get_options()['config_file'] )

        req.content_type = 'text/html'
        req.send_http_header()
        self.fs = util.FieldStorage(req,keep_blank_values=True)

        try:
            self.stats = get_stats()
            try:
                req.write( self.header() )
            except:
                req.write( html_header (self.title) )
        
            if len(self.fs) == 0: # before user submission
                req.write(html_preamble())
                self.form(req)
                req.write(html_help(self.help))

            else: # process user submission
                check_ip(req.get_remote_host(apache.REMOTE_NOLOOKUP))  # make sure user is not spamming
                msg = self.process()
                # if we get this far: Succes!
                req.write(html_success(msg))
                req.write(html_back())

        except InputError as ie: # Invalid user input
            req.write(html_error(str(ie)))
            req.write(html_back())
        
        except SecurityError as he: # User has been permanently or temporarily banned
            req.write(html_error(str(he)))
            req.write(html_back())
        
        except MixError as me: # Error interacting with mixmaster process
            req.write(html_error(str(me)))
            req.write(html_back())

        except ConfigError as ce: # Error parsing configuration
            req.write(html_error(str(ce)))
            req.write(html_back())

        except Exception as ex: # Catch all remaining errors
            req.write('<h2>Unhandled Error!</h2>')
            req.write('<p>'+str(ex)+'</p>')
            req.write(html_back())

        try:
            req.write( html_from_file(conf.footer) )
        except:
            req.write( html_copyright() )

        req.write('\n\n</body>\n\n</html>\n')

        return apache.OK




###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

from config import conf
from stats import uptime_sort, format_stats

NONE = ''
RANDOM = 'RANDOM'


def html_from_file(file):
    with open(file,'r') as f:
        return f.read()


def html_header(title):
    return '''<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">

<head>
  <title>'''+title+'''</title>
</head>

<body>

<h1>'''+title+'''</h1>


'''


def html_copyright():
    return '''
<div id="copyright">
<a href="http://pyanon.sourceforge.net">Pyano</a> - &copy; Sean Whitbeck 2010
</div>
'''


def html_error(msg):
    return '''<h2>Error!</h2>

<p>
'''+msg+'''
</p>
'''


def html_preamble():
    return '''<p>
Use this mixmaster web interface to send anonymous emails. If no remailers are selected
in the chain, then mixmaster will automatically select a remailer chain for you. If you intend
on using the <em>From</em> field please read the instructions <a href="#from">here</a>. For
more detailed instructions, refer to the <a href="#help">help</a> section below.
</p>
<p>
Fields marked with <strong>*</strong> are mandatory.
</p>
'''

def html_help( msg ):
    return '''<h2><a name="help">Help</a></h2>

<p>
The mixmaster interface is a completely anonymous solution. 
Messages sent via this interface cannot even be traced back to matterhorn. 
Use it responsibly. If you spam through it, you risk getting banned.
</p>

<p>
This interface wraps your message in layers of PGP encryption. 
Each remailer in the chain unwraps one layer then sends to the next one. 
In this way no one remailer knows the source, destination, and contents. 
The longer the chain you use, the more guaranteed the anonymity. 
However, this comes at the cost of reliabiltyy. Indeed, remailers can be unreliable 
and there is no guarantee that every message sent will arrive at it's destination. Increasing
the number of copies increases the reliability.
</p>

<p>
'''+msg+''' Choose your remailers. Enter the message body, then click send. 
A sent page will appear to inform you on the outcome. Wait for it!.
</p>

<h3>Stats</h3>

<p>
The remailers are listed in the drop lists with some stats. Those stats are as follows:
</p>
<ul class="description">
  <li><dfn>Name</dfn>: The name of the remailer.</li>
  <li><dfn>Latency</dfn>: Latency means how long the remailer will hold onto the message before delivering it to the next remailer in line.</li>
  <li><dfn>Uptime</dfn>: This is a percentage. Obviously you want to use remailers with high uptime. I would choose 95% or better.</li>
  <li><dfn>Middle</dfn>: If you see this word next to a remailer it means that this remailer only operates in middleman mode. That means you can only use it in the middle of a chain. If you use it at the end, your message will not be delivered.</li>
  <li><dfn>From</dfn>: This remailer may allow user-supplied from headers. See below for more information.</li>
</ul>

<h3><a name="from">From Lines</a></h3>

<p>
Not all remailers support from lines. If you wish 
to use a custom from line rather than one the remailer chooses for you, you must select 
a remailer that supports custom from lines as the last remailer in the chain! Put your 
from line in this format: <kbd>"John Smith" &lt;john@domain.com&gt;</kbd>
</p>

'''
    

def html_chain(stats):
    s = ''
    for i in range(0, conf.chain_max_length):
        s += '''      <select name="chain'''+str(i)+'''" >
        <option value="'''+NONE+'''" selected="selected"></option>
        <option value="'''+RANDOM+'''">Random</option>
'''
        for remailer in uptime_sort(stats):
            s += '        <option value="'+remailer+'">'+format_stats(remailer,stats[remailer]).replace(' ',"&nbsp;")+'</option>\n'
        s += '      </select><br/>'
    return s


def html_copies():
    s = '''      <select name="copies">
        <option value="1" selected="selected">1</option>
'''
    for i in range(1, conf.max_copies):
        s += '        <option value="'+str(i+1)+'" >'+str(i+1)+'</option>\n'
    s += '      </select>'
    return s


def html_back():
    return '''<p>
Press your browser's back button to come back to the mixmaster web interface.
</p>
'''

def html_success(msg):
    return '''<h2>Success!</h2>

<p>
'''+msg+'''
</p>
'''



from setuptools import setup, find_packages

setup(name="pyano",
      version = "0.5.2b",
      packages = find_packages(),
      license = "GPLv3",
      author = "Sean Whitbeck",
      author_email = "sean@neush.net",
      description = "Mixmaster web interface for mod_python",
      keywords = 'mixmaster remailer web interface',
      url = "http://pyanon.sourceforge.net",
      zip_safe = True,
      include_package_data = True
)



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

import datetime
from config import conf

datetime_fmt = '%H%M%S%m%d%Y'

class SecurityError(Exception):
    pass

def get_history():
    history = {}
    try:
        with open(conf.hist_file,'r') as f:
            for line in f:
                times = line.strip('\n').split(':')
                ip = times.pop(0)
                history[ip] = [ datetime.datetime.strptime(t,datetime_fmt) for t in times ]
    except IOError: # hist file probably does not exist
        pass
    return history


def prune_history(hist):
    now = datetime.datetime.today()
    cutoff = now - datetime.timedelta(0,0,0,0,conf.hist_window)
    to_del = []
    for ip, times in hist.iteritems():
        new_times = []
        for t in times:
            if t >= cutoff:
                new_times.append(t)
        if new_times: 
            hist[ip] = new_times
        else: # no more recorded uses in the time window: remove all history for that ip
            to_del.append(ip)
    for ip in to_del:
        del hist[ip]
    return hist


def write_history(hist):
    try:
        with open(conf.hist_file,'w') as f:
            for ip, times in hist.iteritems():
                f.write(ip)
                for t in times:
                    f.write( ':' + t.strftime(datetime_fmt) )
                f.write("\n")
    except IOError: # histfile is probably not writable
        pass


def check_hist(ip):
    hist = prune_history( get_history() )
    if ip in hist:
        hist[ip].append( datetime.datetime.today() )
        write_history(hist)
        if len(hist[ip]) > 2*conf.hist_max_uses:
            ban(ip)
            raise SecurityError('Your ip has been banned from using this service. If you feel this is not warranted, please contact this service\'s operator')
        if len(hist[ip]) > conf.hist_max_uses:
            raise SecurityError('In order to prevent abuse of this service, you can only send a limited number of emails at a given time. Please try again later. If you persist in trying to use this interface too frequently, you will eventually get banned from this service.')
    else:
        hist[ip] = [ datetime.datetime.today() ]
        write_history(hist)


def check_banned(ip):
    banned = get_banned()
    if ip in banned:
        raise SecurityError('Your ip has been banned from using this service. If you feel this is not warranted, please contact this service\'s operator')


def ban(ip):
    banned = get_banned()
    banned.append(ip)
    write_banned(banned)


def get_banned():
    banned = []
    try:
        with open(conf.banned_file,'r') as f:
            banned = f.read().split('\n')
    except IOError: # file probably does not exist
        pass
    return banned

def write_banned(banned):
    with open(conf.banned_file,'w') as f:
        f.write('\n'.join(banned))


def check_ip(ip):
    check_banned(ip)
    check_hist(ip)



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

import ConfigParser

class Config:
    pass

conf = Config()

PYANO = 'Pyano'

class ConfigError(Exception):
    pass

def parse_config(config_file):
    config = ConfigParser.SafeConfigParser()
    config.add_section(PYANO)
    # fill defaults
    config.set(PYANO,'mixmaster','/usr/bin/mixmaster')
    config.set(PYANO,'mlist2','')
    config.set(PYANO,'allow_from','')
    config.set(PYANO,'hist_file','')
    config.set(PYANO,'hist_window','15')
    config.set(PYANO,'hist_max_uses','5')
    config.set(PYANO,'banned_file','')
    config.set(PYANO,'chain_max_length','5')
    config.set(PYANO,'max_copies','5')
    config.set(PYANO,'mail_header','')
    config.set(PYANO,'news_header','')
    config.set(PYANO,'footer','')
    config.set(PYANO,'mail2news','mail2news@dizum.com, mail2news@m2n.mixmin.net, mail2news@reece.net.au')
    # read config file
    try:
        config.read(config_file)
        conf.mixmaster = config.get(PYANO,'mixmaster')
        conf.mlist2 = config.get(PYANO,'mlist2')
        conf.allow_from = config.get(PYANO,'allow_from')
        conf.hist_file = config.get(PYANO,'hist_file')
        conf.hist_window = config.getint(PYANO,'hist_window')
        conf.hist_max_uses = config.getint(PYANO,'hist_max_uses')
        conf.banned_file = config.get(PYANO,'banned_file')
        conf.chain_max_length = config.getint(PYANO,'chain_max_length')
        conf.max_copies = config.getint(PYANO,'max_copies')
        conf.mail_header = config.get(PYANO,'mail_header')
        conf.news_header = config.get(PYANO,'news_header')
        conf.footer = config.get(PYANO,'footer')
        conf.mail2news = config.get(PYANO,'mail2news')
    except: # Something is wrong with the configuration file
        raise ConfigError('Error parsing configuration file.')
        




###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

import re
import random

from stats import UPTIME, MIDDLE, FROM
from html import NONE, RANDOM
from config import conf


class InputError(Exception):
    pass

def val_email_local(local):
    ac = "[a-zA-Z0-9_!#\$%&'*+\-=?\^`{|}~]"
    m = re.compile( '^('+ac+'+\.)*'+ac+'+$')
    if not m.match(local):
        raise InputError()


def val_dot_seq(server):
    ac = '[a-zA-Z0-9+\-_]'
    m = re.compile( '^('+ac+'+\.)+'+ac+'+$')
    if not m.match(server):
        raise InputError()


def val_email(addr):
    m = re.compile( '^("?[\w ]*"? *<)?([^>]+)>?$' )
    ok =  m.match(addr)
    try:
        if not ok:
            raise InputError()
        email = ok.group(2)
        parts = email.split('@')
        if len(parts) != 2:
            raise InputError()
        val_email_local(parts[0])
        val_dot_seq(parts[1])
    except InputError:
        raise InputError(addr+' is not a valid email address.')


def val_newsgroups(newsgroups):
    groups = newsgroups.split(',')
    for group in groups:
        val_dot_seq(group)


def val_references(refs):
    m = re.compile( '^<(.*)>$' )
    try:
        for ref in refs.split(' '):
            ok = m.match(ref)
            if not ok:
                raise InputError()
            val_email( ok.group(1) )
    except:
        raise InputError(ref+' is not a valid reference.')


def val_mail2news(mail2news):
    for addr in mail2news.split(','):
        val_email(addr)


def val_hashcash(hc):
    m = re.compile('[a-zA-Z0-9_:+\-=; ]+')
    if not m.match(hc):
        raise InputError(hc+' is not a valid hashcash.')

def val_n_copies(n_copies):
    if n_copies <= 0 or n_copies > conf.max_copies:
        raise InputError('Invalid number of copies.')

# Explanation for the chain parsing:
#
# By default the mixmaster binary will accept messages with duplicate
# remailers (bad for privacy), or a middleman remailer as the final
# remailer (which will silently fail).
#
# On the other hand, when sending multi-part emails, it *requires*
# two random remailers in the chain.
#
# Here we make the following imperfect design choices:
#  - Translate a choice of RANDOM to '*' (the mixmaster random chain)
#    Except if a random remailer is chosen last, in which case this program
#    selects the final remailer and makes sure it is not a middleman.
#  - Check for duplicate remailers. However if the user selects a RANDOM remailer
#    mixmaster MAY select a duplicate remailer. (TODO: read the mixmaster code
#    to make sure...)
#
def parse_chain(fs,stats):
    chain = []
    for i in range(0, conf.chain_max_length):
        remailer = str(fs['chain'+str(i)])
        if remailer == NONE: # empty remailer, we stop the chain
            break
        elif remailer == RANDOM:
            chain.append('*')
        elif remailer in chain: # duplicate remailer!
            raise InputError('Cannot use the same remailer twice in the same chain.')
        elif remailer in stats:
            chain.append( remailer )
        else: # invalid remailer entry
            raise InputError(remailer+' is not a valid remailer.')

    if chain:
        last = chain[len(chain)-1]
        if last == '*': # make sure final remailer is neither a duplicate nor a middleman
            chain[len(chain)-1] = get_random_remailer(stats,isValid=lambda r : not( r in chain or stats[r][MIDDLE]) ) 
        elif stats[last][MIDDLE]: # make sure that final remailer is not a middleman
            raise InputError('Cannot select a "middle" remailer as your final remailer.')
    return chain


def get_random_remailer(stats,isValid=lambda r : True):
    # first select the subset of remailers that have 100% reliability
    reliable = set([])
    for remailer, s in stats.iteritems():
        if s[UPTIME] == 100.0 and isValid(remailer):
            reliable.add(remailer)
    if not reliable: # there are no 100% reliable remailers
        for remailer in stats.itervalues():
            if isValid(remailer):
                reliable.add(remailer)
    if not reliable: # no remailers matching criteria
        raise InputError('Could not select adequate random remailer.')
    # get a random remailer
    return random.choice(list(reliable))
